// ----------------------------------------------------------------------------
//
// OpenSteer Daemon
//
// Copyright (c) 2009
// Original author: Thomas Krijnen
//
// Based on: OpenSteer -- Steering Behaviors for Autonomous Characters
//   Copyright (c) 2002-2005, Sony Computer Entertainment America
//   Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
//
// ----------------------------------------------------------------------------


#ifdef _WIN32
	#include <winsock2.h>
	#pragma comment(lib, "Ws2_32.lib")
	#pragma comment(lib, "user32")	
#else
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <fcntl.h>
#endif

#include "OpenSteer/OpenSteerDemo.h"
#include "OpenSteer/Draw.h"
#include "OpenSteer/SimpleVehicle.h"
#include <cstring>
#include <sstream>
#include <map>
#include <cstdlib>

// Socket communication definitions
static int sock;
static bool requested;
static bool startConnection();
static void closeConnection();
static std::string readLine();
static void sendLine(std::string);
static int readInt();
static float readFloat();

// client request location and rotation of actors
// returns per actor: SerialNo,LocX,LocY,LocZ,RotY\n
// returns per actor: SerialNo,LocX,LocY,LocZ,FwdX,FwdY,FwdZ\n
void getSimData(void) {							
	sendLine("okget");						
	sendLine("startactors");
	OpenSteer::AVGroup vehicles = OpenSteer::OpenSteerDemo::selectedPlugIn->allVehicles();
	typedef std::vector<OpenSteer::AbstractVehicle*>::iterator iterator;
	for (iterator i = vehicles.begin(); i!=vehicles.end(); i++){								
		OpenSteer::Vec3 fwd = (**i).forward();
		OpenSteer::Vec3 pos = (**i).position();
		OpenSteer::SimpleVehicle* sv = dynamic_cast<OpenSteer::SimpleVehicle*>(*i);
		std::ostringstream oss;				
		//oss << sv->serialNumber <<  "," << pos.x << "," << pos.y << "," << pos.z << "," << atan2(fwd.z,fwd.x);
		oss << sv->serialNumber <<  "," << pos.x << "," << pos.y << "," << pos.z << "," << fwd.x << "," << fwd.y << "," << fwd.z;
		sendLine(oss.str());
	}
	sendLine("endactors");
	requested = true;
}

// are we up and running?
void ping(void) {								
	std::cout << " +ping\n";
	sendLine("okping");
}

// reset the plugin
void reset(void) {								
	std::cout << " +reset\n";
	OpenSteer::OpenSteerDemo::selectedPlugIn->reset();
	sendLine("okreset");
}

// store these for later use, when entering draw loop later
static int _argc;								
static char **_argv;

// launch the OpenGL display window and stop network communication
void showRender(void) {
	sendLine("okrender");
	OpenSteer::NetworkPlugIn* np = dynamic_cast<OpenSteer::NetworkPlugIn*>(OpenSteer::OpenSteerDemo::selectedPlugIn);
	np->canDraw = true;
	OpenSteer::enableAnnotation = true;
	std::cout << " +initializeGraphics\n";
	OpenSteer::initializeGraphics (_argc, _argv);
	std::cout << " +runGraphics\n";
	OpenSteer::runGraphics ();
}

// terminate daemon
void quit(void) {
	sendLine("okquit");
	exit(0);
}

// map of global network commands
static std::map<std::string, void(*)(void)> x;	

// check and process event, called from main loop
static bool processNetworkEvents(void) {		
	requested = false;							
	std::string command = readLine();
	if ( command != "" ) {
	if(x.find(command)!=x.end()) {x[command](); return true;}	// global command
	else {										
		std::cout << " +" << command;
		OpenSteer::PlugIn *p = OpenSteer::OpenSteerDemo::selectedPlugIn;
		OpenSteer::NetworkPlugIn* np = dynamic_cast<OpenSteer::NetworkPlugIn*>(p);
		command = command.substr(0,command.length()-1);
		std::string args = np->hasCommand(command);				// plugin command
		if ( args == "na" ) {
			std::cout << "command unknown\n";
		} else {
			std::string::iterator it;
			int count = args.length();							// read arguments from socket
			float* params = new float[count];
			int i = 0;
			std::cout << " | ";
			for( it = args.begin(); it != args.end(); it++ ) {
				if (*it == 'f') {
					float arg = readFloat();
					params[i++] = arg;
				}
				if (*it == 'i') {
					int arg = readInt();
					params[i++] = *((float*)&arg);
				}
			}
			std::cout << std::endl;
			if (np->processCommand((void*)params) ) {
				sendLine("ok"+command);							// confirm to client
			}
			
		}
		return true;
	}
	}
	return false;
}

int main (int argc, char **argv) 
{
	_argc = argc; _argv = argv;
	x["get\n"] = &getSimData;
	x["render\n"] = &showRender;
	x["quit\n"] = &quit;
	x["ping\n"] = &ping;
	x["reset\n"] = &reset;
	OpenSteer::OpenSteerDemo::initialize ();
	OpenSteer::enableAnnotation = false;

	// we're using animation mode so differences in realised frame rate
	// by the client application don't get recorded as leaps in the simulation
	OpenSteer::OpenSteerDemo::clock.setAnimationMode(true);
	OpenSteer::OpenSteerDemo::clock.setFixedFrameRate(10);
	
	if ( ! startConnection() ) {return false;}

	while(1) {
		processNetworkEvents();
		if (requested) {
			// update only if position of actors has been requested
			OpenSteer::OpenSteerDemo::clock.advanceSimulationTimeOneFrame();
			OpenSteer::OpenSteerDemo::updateSimulationAndRedraw ();
		}
#ifdef _WIN32
		Sleep(2);
#else
		usleep(2000);
#endif
	}

	closeConnection();

	return EXIT_SUCCESS;
}

// -------------------
// boring socket stuff
// -------------------

static bool startConnection() {
#ifdef _WIN32
	WSADATA wsa;
	if (WSAStartup(MAKEWORD(2,0), &wsa) != NO_ERROR) {return false;}
#endif
	sockaddr_in address;
	memset(&address, 0, sizeof(sockaddr_in));
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = htonl(INADDR_ANY);
	address.sin_port = htons(20002);
	int s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (bind(s, (struct sockaddr *)&address, sizeof(sockaddr_in)) < 0 ) {
		std::cout << "Unable to bind socket" << std::endl; 
		return false;
	}
	std::cout << "Waiting for connection..." << std::endl;
	listen(s,1);
	sock = accept(s,0,0);
	std::cout << "\rConnected!               \n";
	if (sock < 0 ) {
		std::cout << "Invalid Socket" << std::endl; 
		return false;
	}
#ifdef _WIN32
	closesocket(s);
	u_long arg = 1;
	ioctlsocket(sock, FIONBIO, &arg);
#else
	close(s);
	int nonBlocking = 1;
    fcntl(sock, F_SETFL, O_NONBLOCK, nonBlocking );
#endif
	return true;

}

static void closeConnection() {
#ifdef _WIN32
	closesocket(sock);
	WSACleanup();
#else
	close(sock);
#endif
}

static std::string readLine() {
	std::string ret;
	while (1) {
		char c;
		const int r = recv(sock, &c, 1, 0);
		if ( r < 1 ) {return "";}
		ret += c;
		if ( c == '\n' ) {return ret;}
	}
}

static void sendLine(std::string s) {
	s += "\n";
	send(sock,s.c_str(),s.length(),0);
}

// read a line from the networksocket and convert to float
static float readFloat() {
	std::string arg;
	while ( arg == "" ) {
		arg = readLine();
	}
	const char *c_ptr=arg.c_str();
	float f = (float) atof(c_ptr);
	std::cout << f << " ";
	return f;
}

// read a line from the networksocket and convert to int
static int readInt() {
	std::string arg;
		while ( arg == "" ) {							
		arg = readLine();
	}
	const char *c_ptr=arg.c_str();
	int i = (int) atoi(c_ptr);
	std::cout << i << " ";
	return i;
}
